/**************************************************
** (shared) library funcs for Builder Calculator
**
***************************************************/

#include <elfapi.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Procedure to make shared library:
 *
 * SUNOS:
 *        cc -DSUNOS=1 -I<applix_exec_dir> -pic -c calc.c
 *        ld -o libcalc.so -assert pure-text calc.o
 *
 * SOLARIS:
 *        cc -DSOLARIS=1 -K PIC -G -I<applix_exec_dir> -o libcalc.so calc.c
 *
 * HP800:
 *        cc -DHP800=1 +z -I<applix_exec_dir> -c calc.c
 *        ld -o libcalc.so -E -b calc.o
 *
 * IBM:
 *
 * NT:
 *        cl -DAXMS=1 -DUSE_PROTOS -I<applix_exec_dir> -c calc.c
 *        lib -machine:i386 -out:calc.lib -def:calc.def calc.obj
 *        link -dll -entry:_CRT_INIT@12 -out:libcalc.dll calc.obj calc.exp \
 *                  d:\ax\axexec\axdata\elf\elfapi.lib d:\ax\axobj\axdata\rts\axmain.lib \
 *                  libc.lib kernel32.lib user32.lib gdi32.lib winspool.lib comdlg32.lib \
 *                  wsock32.lib advapi32.lib oldnames.lib
 *
 */

/*
 * Functions used from math lib:
 *
 * sin, cos, tan, asin, acos, atan, sinh, cosh, tanh, asinh, acosh, atanh,
 * log, log10, exp, pow, rand
 */

/*
 * general purpose defs
 */
#define streq(s1,s2)	(!strcmp(s1,s2))

#ifndef M_PI
#define	M_PI	3.14159265358979323846
#endif

#ifndef M_E
#define	M_E		2.7182818284590452354
#endif

#if SOLARIS || SYSV4 || PMAX || HP800 || RIOS || ALPHA || AXMS || M_88K || CLIX || SCOUNIX || SGI
#define bcopy(a,b,s) memmove(b,a,s)
#endif


/*
 * sub-function codes 
 */
#include "calcfunc.h"

/*
 * boolean def
 */
enum boolean {
	FALSE,
	TRUE
};

/*
 * trigonometric mode
 */
enum e_trig_mode {
	TRIG_DEGREES,		/* degrees */
	TRIG_GRADIENTS,		/* gradients */
	TRIG_RADIANS		/* radians */
};

/*
 * display mode
 */
enum e_display_mode {
	DISPLAY_ENG,		/* engineering mode */
	DISPLAY_FIX,		/* fixed point mode */
	DISPLAY_SCI			/* scientific mode */
};

/*
 * base radix
 */
enum e_base_radix {
	BASE_BINARY = 2,
	BASE_OCTAL = 8,
	BASE_DECIMAL = 10,
	BASE_HEX = 16
};

/*
 * unsigned mask
 */
enum e_unsigned_mask {
	MASK_CHAR,
	MASK_WORD,
	MASK_DWORD
};

/*
 * parentheses open/close codes
 */
enum e_paren_type {
	CALC_OPEN_PAREN,
	CALC_CLOSE_PAREN
};

/*
 * node to be pushed/popped on/from stack when parenthesis typed
 */
typedef struct paren_stack_t {
	double cur_number;
	char cur_operator[8];
} paren_stack_t;

/*
 * state of calculator
 */
typedef struct calc_state {
	double cur_number;					/* current number that is being displayed */
	double prev_number;					/* previous number remembered for binary op */
	char cur_operator[8];				/* current binary op */
	char typing_string[128];			/* number that is being typed */
	int typing_dex;						/* next typed char goes here in typing_string[] */
	enum boolean seen_dot;				/* TRUE if decimal point has already typed in */
	enum boolean doing_exp;				/* TRUE if exponential portion being typed in */
	int precision;						/* number of decimal places in FIX mode */
	int accuracy;						/* number of decimal places in ENG mode */
	enum boolean in_typing;				/* TRUE if in typing */
	enum boolean inv_func;				/* TRUE if INV on */
	enum boolean hyp_func;				/* hyperbolic functions */
	enum e_display_mode display_mode;	/* DISPLAY_ENG, DISPLAY_FIX, DISPLAY_SCI */
	enum e_trig_mode trig_mode;			/* TRIG_DEGREES, TRIG_GRADIENTS, TRIG_RADIANS */
	enum e_base_radix base_radix;		/* BASE_DECIMAL, BASE_HEX, BASE_OCTAL, BASE_BINARY */
	enum e_unsigned_mask unsigned_mask;	/* MASK_CHAR, MASK_WORD, MASK_DWORD */
	int paren_stack_dex;				/* current depth of parenthesis stack */
#define CALC_MAX_PARENS		16
	paren_stack_t paren_stack[CALC_MAX_PARENS];	/* parenthesis stack */
	int num_memory_regs;				/* number of memory registers */
	double *memory_regs;				/* array of registers - variable size */
} calc_state;

/*
 * definition of node in function dispatch table
 */
typedef struct func_table_t {
	int calc_func_code;
	elfData (*calc_funcP) AXP_((calc_state **, elfData));
} func_table_t;

static char error_msg[256];


#define ALMOST_ZERO			(0.0000000000001)
#define is_almost_zero(d)	calc_double_equal(ALMOST_ZERO,d)

/***************************************************************************
function: calc_double_equal (d1, d2) - returns true if the absolute difference btw. two
numbers is less than or equal to ALMOST_ZERO.
*/
static enum boolean
calc_double_equal (d1, d2)
double d1, d2;
{
	double fudge;

	fudge = fabs(d1 * ALMOST_ZERO) + ALMOST_ZERO;
	d1 = fabs(d1 - d2);
	if ( d1 <= fudge )
		return(TRUE);
	else
		return(FALSE);
}


/***************************************************************************
function: calc_reset_input_state - reset calculator input state
*/
static void
calc_reset_input_state (cs)
calc_state **cs;
{
	(*cs)->in_typing = FALSE;
	(*cs)->cur_number = 0.0;
	(*cs)->prev_number = 0.0;
	(*cs)->cur_operator[0] = 0;
	(*cs)->inv_func = FALSE;
	(*cs)->hyp_func = FALSE;
	(*cs)->doing_exp = FALSE;
	(*cs)->seen_dot = FALSE;
	(*cs)->typing_dex = 0;
	(*cs)->paren_stack_dex = 0;

	error_msg[0] = 0;
}

#define CALC_INITIAL_MEM_REGS 1

/***************************************************************************
function: calc_initialize - initialize calculator
*/
static elfData
calc_initialize (cs, argData)
calc_state **cs;
elfData argData;
{
	int size;

	size = sizeof(calc_state);
	*cs = (calc_state *)malloc(size);
	memset(*cs, 0, size);

	calc_reset_input_state(cs);

	(*cs)->display_mode = DISPLAY_ENG;
	(*cs)->base_radix = BASE_DECIMAL;
	(*cs)->trig_mode = TRIG_DEGREES;
	(*cs)->precision = 2;
	(*cs)->accuracy = 10;

	(*cs)->num_memory_regs = CALC_INITIAL_MEM_REGS;		/* initial number of memory registers */
	size = CALC_INITIAL_MEM_REGS * sizeof(double);
	(*cs)->memory_regs = (double*)malloc(size);
	memset((*cs)->memory_regs, 0, size);

	(*cs)->unsigned_mask = MASK_DWORD;

	return(NULL);
}


/***************************************************************************
function: calc_terminate - terminate calculator
*/
static elfData
calc_terminate (cs, argData)
calc_state **cs;
elfData argData;
{
	free((char *)(*cs)->memory_regs);
	free((char *)*cs);
	*cs = NULL;

	return(NULL);
}


/***************************************************************************
function: calc_format_binary_number - format given number in binary representation
*/
static char *
calc_format_binary_number (n)
unsigned int n;
{
	int ix;
	char t[16];
	static char binary_str[41];

#if BINARY_WITHOUT_SPACES
	binary_str[40] = 0;
	for (ix=39; n && ix >= 0; --ix, n >>= 1)
		binary_str[ix] = '0' + (n & 1);
	ix++;
	if (0 == binary_str[ix])	/* if n == 0 */
		binary_str[--ix] = '0';
	return(binary_str+ix);
		
#else	/* WITH_SPACES */
	static char *bs[] = {
		"0000", "0001", "0010", "0011",
		"0100", "0101", "0110", "0111",
		"1000", "1001", "1010", "1011",
		"1100", "1101", "1110", "1111"
	};

	sprintf(t, "%x", n);
	memset(binary_str,' ',40);
	binary_str[40] = 0;

	for (ix=0; t[ix]; ix++)
		strncpy(binary_str+ix*5, bs[(t[ix] <= '9') ? (t[ix]-'0') : (t[ix]-'a'+10)], 4);

	/* trim both ends */
	ix = 40;
	while (--ix >= 0 && ' ' == binary_str[ix]);
	binary_str[ix+1] = 0;

	for (ix = 0; ix < 40 && binary_str[ix] && binary_str[ix] != '1'; ix++);
	
	/* keep at least one 0 */
	if (0 == binary_str[ix])
		binary_str[--ix] = '0';

	return(binary_str+ix);
#endif	
}

/***************************************************************************
function: calc_format_unsigned - format any unsigned number to a given base radix
*/
static char *
calc_format_unsigned (n, base)
unsigned int n;
{
	static char result_str[64];
	
	result_str[0] = 0;
	n = n & 0xFFFFFFFF;
	if (BASE_HEX == base)
		sprintf(result_str, "%X", n);
	else if (BASE_OCTAL == base)
		sprintf(result_str, "%o", n);
	else if (BASE_BINARY == base)
		sprintf(result_str, "%s", calc_format_binary_number(n));
	return(result_str);
}

/***************************************************************************
function: calc_format_result - format "number" according to display mode
*/
static char *
calc_format_result (cs, prefix, number)
calc_state **cs;
char *prefix;
double number;
{
	unsigned int unum;
	char format_str[20];
	static char result[256];
	
	if (BASE_DECIMAL == (*cs)->base_radix) {
		if (DISPLAY_FIX == (*cs)->display_mode) {
			sprintf(format_str, "%%s%%.%df", (*cs)->precision);
			sprintf(result, format_str, prefix, number);
		}
		else if (DISPLAY_ENG == (*cs)->display_mode) {
			sprintf(format_str, "%%s%%.%dG", (*cs)->accuracy);
			sprintf(result, format_str, prefix, number);
		}
		else if (DISPLAY_SCI == (*cs)->display_mode) {
			sprintf(format_str, "%%s%%.%dE", (*cs)->precision);
			sprintf(result, format_str, prefix, number);
		}
	}
	else {
		unum = ((unsigned int)number);
		if ((*cs)->unsigned_mask == MASK_CHAR)
			unum &= 0xFF;
		else if ((*cs)->unsigned_mask == MASK_WORD)
			unum &= 0xFFFF;
		else
			unum &= 0xFFFFFFFF;
		sprintf(result,"%s%s", prefix, calc_format_unsigned(unum, (*cs)->base_radix));
	}
	return(result);
}


/***************************************************************************
function: calc_read_binary - read number from input string in binary representaion
*/
static unsigned int
calc_read_binary (instr)
char *instr;
{
	unsigned int unum;

	if (!instr)
		return (0);
	unum = 0;
	while (*instr) {
		if ('0' == *instr || '1' == *instr) {
			unum <<= 1;
			if ('1' == *instr)
				unum++;
		}
		instr++;
	}
	return (unum);
}

/***************************************************************************
function: calc_read_input_number - read number from input string for given base radix
*/
static double
calc_read_input_number (cs)
calc_state **cs;
{
	unsigned int unum;
	double result, exp_val;
	char instr[128], *t;
	
	result = 0.0;
	strcpy(instr, (*cs)->typing_string);
	if (!instr[0])
		return(result);

	if (BASE_DECIMAL == (*cs)->base_radix) {
	base_10:
		exp_val = 1.0;
		if ((*cs)->doing_exp && (t = strchr(instr, 'E'))) {
			*t++ = 0;
			if (*t)
				exp_val = pow(10.0, atof(t));
		}
		result = atof(instr) * exp_val;
		return(result);
	}
	else if (BASE_HEX == (*cs)->base_radix) {
		sscanf(instr,"%x",&unum);
		result = unum;
	}
	else if (BASE_OCTAL == (*cs)->base_radix) {
		sscanf(instr,"%o",&unum);
		result = unum;
	}
	else if (BASE_BINARY == (*cs)->base_radix) {
		result = calc_read_binary(instr);
	}
	else /* if ((*cs)->base_radix == BASE_DECIMAL) */
		goto base_10;
	return(result);
}

/***************************************************************************
function: calc_input_alphanum - input alpha numeric data
*/
static elfData
calc_input_alphanum (cs, argData)
calc_state **cs;
elfData argData;
{
	char *typed_char;
	int num, char_val;
	
	typed_char = AxStrFromArray(argData, 2);

	/* if not in_typing mode, start now */
	if (!(*cs)->in_typing) {
		(*cs)->cur_number = 0.0;
		(*cs)->typing_dex = 0;
		(*cs)->seen_dot = FALSE;
		(*cs)->doing_exp = FALSE;
	}
	
	/* handle backspaces */
	if (streq(typed_char, "backsp")) {
		if ((*cs)->in_typing) {
			if ((*cs)->typing_dex > 0)
				--(*cs)->typing_dex;
			if ((*cs)->seen_dot && (*cs)->typing_string[(*cs)->typing_dex] == '.')
				(*cs)->seen_dot = FALSE;
			else if ((*cs)->doing_exp && (*cs)->typing_string[(*cs)->typing_dex] == 'E')
				(*cs)->doing_exp = FALSE;
		}
		goto done;
	}

	(*cs)->in_typing = TRUE;

	if (40 < (*cs)->typing_dex)		/* limit to 40 */
		goto done;
	
	/* exponential key */
	if (streq(typed_char, "exp") && BASE_DECIMAL == (*cs)->base_radix && !(*cs)->doing_exp) {
		(*cs)->doing_exp = TRUE;
		(*cs)->typing_string[(*cs)->typing_dex++] = 'E';
		(*cs)->typing_string[(*cs)->typing_dex++] = '+';
	}

	/* digits and A-F */
	char_val = -1;
	if (1 == strlen(typed_char)) {
		if (typed_char[0] >= '0' && typed_char[0] <= '9')
			char_val = typed_char[0] - '0';
		else if (typed_char[0] >= 'A' && typed_char[0] <= 'F')
			char_val = typed_char[0] - 'A' + 10;
		else if (typed_char[0] >= 'a' && typed_char[0] <= 'f')
			char_val = typed_char[0] - 'a' + 10;
	}
			
	if (char_val >= 0 && char_val < (*cs)->base_radix)
		(*cs)->typing_string[(*cs)->typing_dex++] = typed_char[0];
	else if ('.' == typed_char[0] && BASE_DECIMAL == (*cs)->base_radix &&
			 !(*cs)->seen_dot && !(*cs)->doing_exp) {
		(*cs)->seen_dot = TRUE;
		(*cs)->typing_string[(*cs)->typing_dex++] = typed_char[0];
	}

 done:
	(*cs)->typing_string[(*cs)->typing_dex] = 0;
	(*cs)->cur_number = calc_read_input_number(cs);		/* read number from string */

	return(NULL);
}


/***************************************************************************
function: calc_perform_binary_operation - do the simple arithmetic operation on num1 and num2
*/
static double
calc_perform_binary_operation (cs, operator, num1, num2)
calc_state **cs;
char *operator;			/* binary operator */
double num1, num2;		/* two operands */
{
	double result = num2;
	
	if (streq(operator, "pow")) {
		if ((*cs)->inv_func)
			result = pow(num1, 1.0/num2);
		else
			result = pow(num1, num2);
		(*cs)->inv_func = FALSE;
	}
	else {
		switch (operator[0]) {
		case '+':
			result = num1 + num2;
			break;
		case '-':
			result = num1 - num2;
			break;
		case '*':
			result = num1 * num2;
			break;
		case '/':
			result = num1 / num2;
			break;
		case '|':
			result = ((unsigned int)num1) | ((unsigned int)num2);
			break;
		case '&':
			result = ((unsigned int)num1) & ((unsigned int)num2);
			break;
		case '^':
			result = ((unsigned int)num1) ^ ((unsigned int)num2);
			break;
		case '%':
			result = ((unsigned int)num1) % ((unsigned int)num2);
			break;
		default:
			result = num2;
			break;
		}
	}

	return(result);
}


/***************************************************************************
function: calc_push_or_pop_number (cs, paren) - push or pop node from paren stack
*/
static double
calc_push_or_pop_number (cs, paren_code)
calc_state **cs;
enum e_paren_type paren_code;
{
	double result;
	
	result = (*cs)->cur_number;
	if (paren_code == CALC_OPEN_PAREN && (*cs)->paren_stack_dex < CALC_MAX_PARENS) {
		/* push current number and current operator onto stack */
		(*cs)->paren_stack[(*cs)->paren_stack_dex].cur_number = (*cs)->cur_number;
		strcpy((*cs)->paren_stack[(*cs)->paren_stack_dex].cur_operator, (*cs)->cur_operator);
		++(*cs)->paren_stack_dex;
	}
	else if (paren_code == CALC_CLOSE_PAREN && (*cs)->paren_stack_dex > 0) {
		/* pop number and operator from stack and perform the operation */
		--(*cs)->paren_stack_dex;
		if ((*cs)->paren_stack[(*cs)->paren_stack_dex].cur_operator[0])
			result = calc_perform_binary_operation(cs,
								(*cs)->paren_stack[(*cs)->paren_stack_dex].cur_operator,
								(*cs)->paren_stack[(*cs)->paren_stack_dex].cur_number,
								(*cs)->cur_number);
	}
	return(result);
}


/***************************************************************************
function: calc_input_operator (cs, argData) - input arithmetic operator
*/
static elfData
calc_input_operator (cs, argData)
calc_state **cs;
elfData argData;
{
	char *operator;
	
	(*cs)->in_typing = FALSE;
	operator = AxStrFromArray(argData, 2);

	if (operator[0] == '%' && BASE_DECIMAL == (*cs)->base_radix) {
		(*cs)->cur_number = (*cs)->prev_number * (*cs)->cur_number / 100.0;
		return(NULL);
	}

	if (operator[0] == '(' || operator[0] == '[') {
		(*cs)->cur_number = calc_push_or_pop_number(cs, CALC_OPEN_PAREN);
		(*cs)->cur_operator[0] = 0;
		return(NULL);
	}
		
	if ((*cs)->cur_operator[0])
		(*cs)->cur_number = calc_perform_binary_operation(cs, (*cs)->cur_operator,
												(*cs)->prev_number, (*cs)->cur_number);
	(*cs)->prev_number = (*cs)->cur_number;
	strcpy((*cs)->cur_operator, operator);

	if ('=' == (*cs)->cur_operator[0]) {		/* clear operator and prev_num */
		(*cs)->cur_operator[0] = 0;
		(*cs)->prev_number = 0.0;
	}

	if (operator[0] == ']' || operator[0] == ')') {
		(*cs)->cur_number = calc_push_or_pop_number(cs, CALC_CLOSE_PAREN);
		(*cs)->cur_operator[0] = 0;
	}

	return(NULL);
}


/***************************************************************************
function: calc_hyperbolic_func - compute hyperbolic funcs
*/
static double
calc_hyperbolic_func (cs, trig_funcP, num)
calc_state **cs;
double (*trig_funcP) AXP_((double));
double num;
{
	if ((*cs)->inv_func) {
#if !AXMS		/* not available on NT */
		if (sin == trig_funcP)
			trig_funcP = asinh;
		else if (cos == trig_funcP)
			trig_funcP = acosh;
		else if (tan == trig_funcP)
			trig_funcP = atanh;
#endif
	}
	else {
		if (sin == trig_funcP)
			trig_funcP = sinh;
		else if (cos == trig_funcP)
			trig_funcP = cosh;
		else if (tan == trig_funcP)
			trig_funcP = tanh;
	}
	num = (*trig_funcP)(num);
	return(num);
}

/***************************************************************************
function: calc_trig_func - compute trigonometric funcs
*/
static double
calc_trig_func (cs, trig_funcP, num)
calc_state **cs;
double (*trig_funcP) AXP_((double));
double num;
{
	double rads;

	if ((*cs)->hyp_func)
		num = calc_hyperbolic_func(cs, trig_funcP, num);
	else if ((*cs)->inv_func) {
		if (sin == trig_funcP)
			trig_funcP = asin;
		else if (cos == trig_funcP)
			trig_funcP = acos;
		else if (tan == trig_funcP)
			trig_funcP = atan;
		
		rads = (*trig_funcP)(num);
		if (TRIG_DEGREES == (*cs)->trig_mode)
			num = rads * 180.0 / M_PI;
		else if (TRIG_GRADIENTS == (*cs)->trig_mode)
			num = rads * 200.0 / M_PI;
		else
			num = rads;
	}
	else {
		if (TRIG_DEGREES == (*cs)->trig_mode)
			rads = num * M_PI / 180.0;
		else if (TRIG_GRADIENTS == (*cs)->trig_mode)
			rads = num * M_PI / 200.0;
		else
			rads = num;
		num = (*trig_funcP)(rads);
	}
	return(num);
}

/***************************************************************************
function: calc_compute_factorial - compute factorial of a number
*/
static int
calc_compute_factorial (num)
int num;
{
	if (num < 0 || num > 16)	/* factorial of 17 is out of range for "int" */
		return(0);
	else if (0 == num || 1 == num)
		return(1);
	else
		return(num * calc_compute_factorial(num - 1));
}



/***************************************************************************
function: calc_fin_func_pv (payment, interest, num_periods) - present value function
*/
static double
calc_fin_func_pv (payment, interest, num_periods)
double payment, interest, num_periods;
{
	double result;

	if (is_almost_zero(interest))
		return(payment*num_periods);

	result = pow((1.0 + interest), -num_periods);
	result = payment * ((1.0 - result) / interest);
	return (result);
}


/***************************************************************************
function: calc_fin_func_fv (payment, interest, num_periods) - future value
*/
static double
calc_fin_func_fv (payment, interest, num_periods)
double payment, interest, num_periods;
{
	double result;

	if (is_almost_zero(interest))
		return(payment*num_periods);

	result = pow((1.0 + interest), num_periods);
	result = payment * ((result - 1.0) / interest);
	return (result);
}


/***************************************************************************
function: calc_fin_func_pmt (principal, interest, num_periods) - payment function
*/
static double
calc_fin_func_pmt (principal, interest, num_periods)
double principal, interest, num_periods;
{
	double result;

	if (is_almost_zero(num_periods))
		return(principal);

	if (is_almost_zero(interest))
		return(principal/num_periods);

	result = pow((1.0 + interest), -num_periods);
	result = principal * (interest / (1.0 - result));
	return (result);
}


/***************************************************************************
function: calc_fin_func_pmtdue (principal, interest, num_periods) - payment due
*/
static double
calc_fin_func_pmtdue (principal, interest, num_periods)
double principal, interest, num_periods;
{
	double result;

	if (is_almost_zero(num_periods))
		return(principal);
	if (is_almost_zero(interest))
		return(principal/num_periods);

	result = pow((1.0 + interest), -num_periods);
	result = principal * (interest / (1.0 - result)) / (1.0 + interest);
	return (result);
}


/***************************************************************************
function: calc_fin_func_rate (fv, pv, num_periods) - interest rate
*/
static double
calc_fin_func_rate (fv, pv, num_periods)
double fv, pv, num_periods;
{
	double result;
	
	if (is_almost_zero(pv))
		return(0.0);
	if (is_almost_zero(num_periods))
		return(0.0);
	result = pow((fv/pv), (1.0/num_periods));
	result = result - 1.0;
	return(result);
}


/***************************************************************************
function: calc_fin_func_term (pmt, fv, interest) - number of periods (term)
*/
static double
calc_fin_func_term (pmt, fv, interest)
double pmt, fv, interest;
{
	double result;

	if (is_almost_zero(pmt))
		return(0.0);
	if (interest <= -1.0)
		return(0.0);
	result = log(1 + (fv*interest)/pmt) / log(1+interest);
	return(result);
}


/***************************************************************************
function: calc_fin_func_ctrm (pv, fv, interest) - compounded term
*/
static double
calc_fin_func_ctrm (pv, fv, interest)
double pv, fv, interest;
{
	double result;

	if (is_almost_zero(fv))
		return(0.0);
	if (is_almost_zero(pv))
		return(0.0);
	if (is_almost_zero(interest))
		return(0.0);
	result = log(fv/pv) / log(1+interest);
	return(result);
}


/***************************************************************************
function: calc_fin_func_ddb (cost, salvage, life, period, factor) - depreciation
of an asset using double declining balance method
*/
static double
calc_fin_func_ddb (cost, salvage, life, period, factor)
double cost, salvage, life, period, factor;
{
	double result;
	double constant, n;

	if (is_almost_zero(factor))
		factor = 2.0;	/* default value */
	if (is_almost_zero(life))
		return(0.0);
	if (period > life)
		return(0.0);

	result = 0.0;
	constant = factor/life;
	n = log(salvage/cost) / log(1-constant) ;

	if(period <= 1)
		period = 1;

	if(period <= n)
		result = cost * constant * pow(1-constant, period-1);
	else { /* sum until previous period */
		if(period > (n+1))
			result = 0.0;
		else
			result = (cost - salvage) - (cost * (1 - pow(1-constant, period-1)));
	}
	return(result);
}

static double
calc_fin_func_syd (cost, salvage, life, period)
double cost, salvage, life, period;
{
	double result, sigma;

	if (period > life)
		return(0.0);

	sigma = life * ( life + 1)/2;
	if (is_almost_zero(sigma))
		return(0.0);
	result = ((cost - salvage)/sigma) * (life-period+1);
	return(result);
}

static double
calc_fin_func_sln (cost, salvage, life)
double cost, salvage, life;
{
	double result;

	if (is_almost_zero(life))
		return(0.0);
	result = (cost - salvage)/life;
	return(result);
}



/***************************************************************************
function: calc_input_function (argData) - input function name
*/
static elfData
calc_input_function (cs, argData)
calc_state **cs;
elfData argData;
{
	unsigned int unum;
	char *func;
	char *t;
	
	func = AxStrFromArray(argData, 2);

	if (streq(func, "x^2")) {
		if ((*cs)->inv_func)
			(*cs)->cur_number = sqrt((*cs)->cur_number);
		else
			(*cs)->cur_number = (*cs)->cur_number * (*cs)->cur_number;
	}

	else if (streq(func, "log")) {
		if ((*cs)->inv_func)
			(*cs)->cur_number = pow(10.0, (*cs)->cur_number);
		else
			(*cs)->cur_number = log10((*cs)->cur_number);
	}
	else if (streq(func, "ln")) {
		if ((*cs)->inv_func)
			(*cs)->cur_number = exp((*cs)->cur_number);
		else
			(*cs)->cur_number = log((*cs)->cur_number);
	}

	else if (streq(func, "sin"))
		(*cs)->cur_number = calc_trig_func(cs, sin, (*cs)->cur_number);
	else if (streq(func, "cos"))
		(*cs)->cur_number = calc_trig_func(cs, cos, (*cs)->cur_number);
	else if (streq(func, "tan"))
		(*cs)->cur_number = calc_trig_func(cs, tan, (*cs)->cur_number);

	else if (streq(func, "1/x")) {
		(*cs)->cur_number = 1.0 / (*cs)->cur_number;
	}

	else if (streq(func, "x!"))
		(*cs)->cur_number = calc_compute_factorial((int)(*cs)->cur_number);

	else if (streq(func, "int"))
		(*cs)->cur_number = (int)(*cs)->cur_number;
	else if (streq(func, "frac"))
		(*cs)->cur_number = (*cs)->cur_number - (int)(*cs)->cur_number;

	else if (streq(func, "+/-")) {
		if ((*cs)->in_typing && (*cs)->doing_exp) {
			t = strchr((*cs)->typing_string, 'E');
			if (t) {
				t++;
				if (*t == '+')
					*t = '-';
				else if (*t == '-')
					*t = '+';
				else if (*t) {
					bcopy(t, t+1, strlen(t));	/* overlap copy */
					*t = '-';
					(*cs)->typing_dex++;
					(*cs)->typing_string[(*cs)->typing_dex] = 0;
				}
			}
			(*cs)->cur_number = calc_read_input_number(cs);
			return(NULL);		/* keep typing */
		}
		else
			(*cs)->cur_number = 0.0 - (*cs)->cur_number;
	}

	else if (streq(func, "!")) {
		unum = (*cs)->cur_number;
		(*cs)->cur_number = ~unum;
	}

	else if (streq(func, ">>"))
		(*cs)->cur_number = ((int)(*cs)->cur_number) >> 1;
	else if (streq(func, "<<"))
		(*cs)->cur_number = ((int)(*cs)->cur_number) << 1;

	else if (streq(func, "pmt")) {
		if ((*cs)->num_memory_regs >= 3)
			(*cs)->cur_number = calc_fin_func_pmt((*cs)->memory_regs[0],
												  (*cs)->memory_regs[1],
												  (*cs)->memory_regs[2]);
	}
	else if (streq(func, "pmtdue")) {
		if ((*cs)->num_memory_regs >= 3)
			(*cs)->cur_number = calc_fin_func_pmtdue((*cs)->memory_regs[0],
													 (*cs)->memory_regs[1],
													 (*cs)->memory_regs[2]);
	}
	else if (streq(func, "fv")) {
		if ((*cs)->num_memory_regs >= 3)
			(*cs)->cur_number = calc_fin_func_fv((*cs)->memory_regs[0],
												 (*cs)->memory_regs[1],
												 (*cs)->memory_regs[2]);
	}
	else if (streq(func, "pv")) {
		if ((*cs)->num_memory_regs >= 3)
			(*cs)->cur_number = calc_fin_func_pv((*cs)->memory_regs[0],
												 (*cs)->memory_regs[1],
												 (*cs)->memory_regs[2]);
	}
	else if (streq(func, "term")) {
		if ((*cs)->num_memory_regs >= 3)
			(*cs)->cur_number = calc_fin_func_term((*cs)->memory_regs[0],
												   (*cs)->memory_regs[1],
												   (*cs)->memory_regs[2]);
	}
	else if (streq(func, "rate")) {
		if ((*cs)->num_memory_regs >= 3)
			(*cs)->cur_number = calc_fin_func_rate((*cs)->memory_regs[0],
												   (*cs)->memory_regs[1],
												   (*cs)->memory_regs[2]);
	}
	else if (streq(func, "ctrm")) {
		if ((*cs)->num_memory_regs >= 3)
			(*cs)->cur_number = calc_fin_func_ctrm((*cs)->memory_regs[0],
												   (*cs)->memory_regs[1],
												   (*cs)->memory_regs[2]);
	}
	else if (streq(func, "ddb")) {
		if ((*cs)->num_memory_regs >= 5)
			(*cs)->cur_number = calc_fin_func_ddb((*cs)->memory_regs[0],
												  (*cs)->memory_regs[1],
												  (*cs)->memory_regs[2],
												  (*cs)->memory_regs[3],
												  (*cs)->memory_regs[4]);
	}
	else if (streq(func, "syd")) {
		if ((*cs)->num_memory_regs >= 4)
			(*cs)->cur_number = calc_fin_func_syd((*cs)->memory_regs[0],
												  (*cs)->memory_regs[1],
												  (*cs)->memory_regs[2],
												  (*cs)->memory_regs[3]);
	}
	else if (streq(func, "sln")) {
		if ((*cs)->num_memory_regs >= 3)
			(*cs)->cur_number = calc_fin_func_sln((*cs)->memory_regs[0],
												  (*cs)->memory_regs[1],
												  (*cs)->memory_regs[2]);
	}

	(*cs)->in_typing = FALSE;
	(*cs)->inv_func = FALSE;
	(*cs)->hyp_func = FALSE;

	return(NULL);
}


/***************************************************************************
function: calc_input_constant (argData) - input a constant number
*/
static elfData
calc_input_constant (cs, argData)
calc_state **cs;
elfData argData;
{
	char *data;
	
	data = AxStrFromArray(argData, 2);

	if (streq(data,"m_pi"))
		(*cs)->cur_number = M_PI;
	else if (streq(data,"m_e"))
		(*cs)->cur_number = M_E;
	else if (streq(data,"rand"))
		(*cs)->cur_number = rand();
	else {
		strcpy((*cs)->typing_string, data);
		(*cs)->cur_number = calc_read_input_number(cs);
	}
	
	return(NULL);
}


/***************************************************************************
function: calc_set_display_mode (argData) - set number display mode
*/
static elfData
calc_set_display_mode (cs, argData)
calc_state **cs;
elfData argData;
{
	char *mode;
	
	mode = AxStrFromArray(argData, 2);

	if (streq(mode, "ENG"))
		(*cs)->display_mode = DISPLAY_ENG;
	else if (streq(mode, "FIX"))
		(*cs)->display_mode = DISPLAY_FIX;
	else if (streq(mode, "SCI"))
		(*cs)->display_mode = DISPLAY_SCI;
	else if (streq(mode, "CHAR"))
		(*cs)->unsigned_mask = MASK_CHAR;
	else if (streq(mode, "WORD"))
		(*cs)->unsigned_mask = MASK_WORD;
	else if (streq(mode, "DWORD"))
		(*cs)->unsigned_mask = MASK_DWORD;
	(*cs)->in_typing = FALSE;

	return(NULL);
}


/***************************************************************************
function: calc_set_base_radix (argData) - set base radix for number system
*/
static elfData
calc_set_base_radix (cs, argData)
calc_state **cs;
elfData argData;
{
	char *mode;
	
	mode = AxStrFromArray(argData, 2);

	if (streq(mode, "10"))
		(*cs)->base_radix = BASE_DECIMAL;
	else if (streq(mode, "16"))
		(*cs)->base_radix = BASE_HEX;
	else if (streq(mode, "8"))
		(*cs)->base_radix = BASE_OCTAL;
	else if (streq(mode, "2"))
		(*cs)->base_radix = BASE_BINARY;

	(*cs)->in_typing = FALSE;
	
	return(NULL);
}


/***************************************************************************
function: calc_get_base_radix (argData) - get base radix for number system
*/
static elfData
calc_get_base_radix (cs, argData)
calc_state **cs;
elfData argData;
{
	return(AxMakeIntData((*cs)->base_radix));
}


/***************************************************************************
function: calc_set_trig_mode (argData) - set mode for trigonometric functions
*/
static elfData
calc_set_trig_mode (cs, argData)
calc_state **cs;
elfData argData;
{
	char *mode;
	
	mode = AxStrFromArray(argData, 2);

	if (streq(mode, "DEG"))
		(*cs)->trig_mode = TRIG_DEGREES;
	else if (streq(mode, "GRA"))
		(*cs)->trig_mode = TRIG_GRADIENTS;
	else if (streq(mode, "RAD"))
		(*cs)->trig_mode = TRIG_RADIANS;

	return(NULL);
}


/***************************************************************************
function: calc_get_display_info (argData) - returns info to be displayed in
the display panel of calculator in the form array of strings
*/
static elfData
calc_get_display_info (cs, argData)
calc_state **cs;
elfData argData;
{
	elfData dispInfo;
	int ix;
	char mode[256];
	
	dispInfo = AxMakeArray(9);

	/* [0] - current number */
	if ((*cs)->in_typing && *(*cs)->typing_string)
		dispInfo = AxAddStrToArray(dispInfo, 0, (*cs)->typing_string);
	else
		dispInfo = AxAddStrToArray(dispInfo, 0, calc_format_result(cs, "", (*cs)->cur_number));

	/* [1] - display mode */
	mode[0] = 0;
	if ((*cs)->base_radix == BASE_DECIMAL) {
		if (DISPLAY_ENG == (*cs)->display_mode)
			strcpy(mode, "ENG");
		else if (DISPLAY_FIX == (*cs)->display_mode)
			strcpy(mode, "FIX");
		else if (DISPLAY_SCI == (*cs)->display_mode)
			strcpy(mode, "SCI");
	}
	else {
		if (MASK_CHAR == (*cs)->unsigned_mask)
			strcpy(mode, "CHAR");
		else if (MASK_WORD == (*cs)->unsigned_mask)
			strcpy(mode, "WORD");
		else if (MASK_DWORD == (*cs)->unsigned_mask)
			strcpy(mode, "DWORD");
	}
	dispInfo = AxAddStrToArray(dispInfo,1, mode);

	/* [2] - base radix */
	mode[0] = 0;
	if (BASE_DECIMAL == (*cs)->base_radix)
		strcpy(mode, "DEC");
	else if (BASE_HEX == (*cs)->base_radix)
		strcpy(mode, "HEX");
	else if (BASE_OCTAL == (*cs)->base_radix)
		strcpy(mode, "OCT");
	else if (BASE_BINARY == (*cs)->base_radix)
		strcpy(mode, "BIN");
	dispInfo = AxAddStrToArray(dispInfo, 2, mode);
	
	/* [3] - trig mode */
	mode[0] = 0;
	if (TRIG_DEGREES == (*cs)->trig_mode)
		strcpy(mode, "DEG");
	else if (TRIG_GRADIENTS == (*cs)->trig_mode)
		strcpy(mode, "GRA");
	else if (TRIG_RADIANS == (*cs)->trig_mode)
		strcpy(mode, "RAD");
	dispInfo = AxAddStrToArray(dispInfo, 3, mode);
	
#if 0
	/* [4] - operator */
	mode[0] = 0;
	if ((*cs)->doing_exp && (*cs)->in_typing)
		strcpy(mode, "exp");
	else if ((*cs)->cur_operator[0])
		strcpy(mode, (*cs)->cur_operator);
	dispInfo = AxAddStrToArray(dispInfo, 4, mode);
#endif

	/* [5] - INV */
	mode[0] = 0;
	if ((*cs)->inv_func)
		strcpy(mode, "INV");
	dispInfo = AxAddStrToArray(dispInfo, 5, mode);

	/* [6] - HYP */
	mode[0] = 0;
	if ((*cs)->hyp_func)
		strcpy(mode, "HYP");
	dispInfo = AxAddStrToArray(dispInfo, 6, mode);

	/* [7] - error message */
	dispInfo = AxAddStrToArray(dispInfo, 7, error_msg);

	/* [8] - parenthesis info */
	mode[0] = 0;
	if ((*cs)->paren_stack_dex > 0) {
		for (ix = 0; ix < (*cs)->paren_stack_dex; ++ix) {
			if ((*cs)->paren_stack[ix].cur_operator[0]) {
				strcat(mode, calc_format_result(cs, "", (*cs)->paren_stack[ix].cur_number));
				strcat(mode, " ");
				strcat(mode, (*cs)->paren_stack[ix].cur_operator);
				strcat(mode, " ");
			}
			strcat(mode, "(");
		}
		if (strlen(mode) > 80)
			mode[80] = 0;
	}

	if ((*cs)->cur_operator[0]) {
		strcat(mode, calc_format_result(cs,"",(*cs)->prev_number));
		strcat(mode, " ");
		strcat(mode, (*cs)->cur_operator);
	}

	dispInfo = AxAddStrToArray(dispInfo, 8, mode);
	return(dispInfo);
}


/***************************************************************************
function: calc_misc_function (argData) - input a misc. function
*/
static elfData
calc_misc_function (cs, argData)
calc_state **cs;
elfData argData;
{
	char *func;
	
	func = AxStrFromArray(argData, 2);
	
	if (streq(func, "clear"))
		calc_reset_input_state(cs);

	else if (streq(func, "drg")) {
		(*cs)->trig_mode++;
		if ((*cs)->trig_mode > TRIG_RADIANS)
			(*cs)->trig_mode = TRIG_DEGREES;
	}
	else if (streq(func, "base")) {
		if (BASE_BINARY == (*cs)->base_radix)
			(*cs)->base_radix = BASE_OCTAL;
		else if (BASE_OCTAL == (*cs)->base_radix)
			(*cs)->base_radix = BASE_DECIMAL;
		else if (BASE_DECIMAL == (*cs)->base_radix)
			(*cs)->base_radix = BASE_HEX;
		else if (BASE_HEX == (*cs)->base_radix)
			(*cs)->base_radix = BASE_BINARY;
		(*cs)->in_typing = FALSE;
	}
	else if (streq(func, "disp")) {
		if ((*cs)->base_radix != BASE_DECIMAL) {
			(*cs)->unsigned_mask++;
			if ((*cs)->unsigned_mask > MASK_DWORD)
				(*cs)->unsigned_mask = MASK_CHAR;
		}
		else {
			(*cs)->display_mode++;
			if ((*cs)->display_mode > DISPLAY_SCI)
				(*cs)->display_mode = DISPLAY_ENG;
		}
		(*cs)->in_typing = FALSE;
	}

	else if (streq(func, "precision"))
		(*cs)->precision = AxIntFromArray(argData, 3);

	else if (streq(func, "accuracy"))
		(*cs)->accuracy = AxIntFromArray(argData, 3);

	else if (streq(func, "inv"))
		(*cs)->inv_func = !(*cs)->inv_func;

	else if (streq(func, "hyp"))
		(*cs)->hyp_func = !(*cs)->hyp_func;

	return(NULL);
}
		

/***************************************************************************
function: calc_get_memory_info (cs, argData) - format memory registers into array of
strings for display.
*/
static elfData
calc_get_memory_info (cs, argData)
calc_state **cs;
elfData argData;
{
	char prefix[64];
	elfData memInfo;
	int ix;
	
	memInfo = AxMakeArray((*cs)->num_memory_regs);
	for (ix = 0; ix < (*cs)->num_memory_regs; ++ix) {
		sprintf(prefix,"Register %3d:     ", ix);
		memInfo = AxAddStrToArray(memInfo, ix,
								  calc_format_result(cs, prefix, (*cs)->memory_regs[ix]));
	}
	return(memInfo);
}
	

/***************************************************************************
function: calc_memory_function (cs, argData) - do memory functions
*/
static elfData
calc_memory_function (cs, argData)
calc_state **cs;
elfData argData;
{
	char *func;
	int memdex;
	int size;
	double exch;
	
	func = AxStrFromArray(argData, 2);
	if (streq(func,"new_mem")) {		/* create new register */
		memdex = ++(*cs)->num_memory_regs;
		size = memdex * sizeof(double);
		(*cs)->memory_regs = (double *) realloc((*cs)->memory_regs, size);
		(*cs)->memory_regs[memdex - 1] = 0.0;
	}
	else if (streq(func, "del_mem")) {	/* delete register */
		memdex = AxIntFromArray(argData, 3);
		if (memdex >= 1 && memdex < (*cs)->num_memory_regs) {
			bcopy((*cs)->memory_regs+memdex+1, (*cs)->memory_regs+memdex, 
				  ((*cs)->num_memory_regs - memdex - 1) * sizeof(double));
			memdex = --(*cs)->num_memory_regs;
			size = (memdex * sizeof(double));
			(*cs)->memory_regs = (double *) realloc((*cs)->memory_regs, size);
		}
	}
	else if (streq(func, "store")) {	/* store current value in selected register */
		memdex = AxIntFromArray(argData, 3);
		if (memdex >= 0 && memdex < (*cs)->num_memory_regs)
			(*cs)->memory_regs[memdex] = (*cs)->cur_number;
	}
	else if (streq(func, "recall")) {	/* recall from selected register */
		memdex = AxIntFromArray(argData, 3);
		if (memdex >= 0 && memdex < (*cs)->num_memory_regs)
			(*cs)->cur_number = (*cs)->memory_regs[memdex];
	}
	else if (streq(func, "addto_mem")) {		/* add current value to mem register and store */
		memdex = AxIntFromArray(argData, 3);
		if (memdex >= 0 && memdex < (*cs)->num_memory_regs) {
			if ((*cs)->inv_func)
				(*cs)->memory_regs[memdex] -= (*cs)->cur_number;
			else
				(*cs)->memory_regs[memdex] += (*cs)->cur_number;
		}
		(*cs)->inv_func = FALSE;
	}
	else if (streq(func, "clear")) {	/* clear current register's contents */
		memdex = AxIntFromArray(argData, 3);
		if (memdex >= 0 && memdex < (*cs)->num_memory_regs)
			(*cs)->memory_regs[memdex] = 0.0;
	}
	else if (streq(func, "exch")) {		/* exchange current value with current mem register */
		memdex = AxIntFromArray(argData, 3);
		if (memdex >= 0 && memdex < (*cs)->num_memory_regs) {
			exch = (*cs)->cur_number;
			(*cs)->cur_number = (*cs)->memory_regs[memdex];
			(*cs)->memory_regs[memdex] = exch;
		}
	}
	(*cs)->in_typing = FALSE;
	return(NULL);
}


/***************************************************************************
 * function dispatch table
 *
 */
static func_table_t calc_func_table[] =
{
	{ CALC_INITIALIZE,			calc_initialize			},
	{ CALC_INPUT_ALPHANUM,		calc_input_alphanum		},
	{ CALC_INPUT_OPERATOR,		calc_input_operator		},
	{ CALC_INPUT_CONSTANT,		calc_input_constant		},
	{ CALC_INPUT_FUNCTION,		calc_input_function		},
	{ CALC_SET_DISP_MODE,		calc_set_display_mode	},
	{ CALC_SET_BASE_RADIX,		calc_set_base_radix		},
	{ CALC_SET_TRIG_MODE,		calc_set_trig_mode		},
	{ CALC_GET_DISP_INFO,		calc_get_display_info	},
	{ CALC_MISC_FUNCTION,		calc_misc_function		},
	{ CALC_GET_BASE_RADIX,		calc_get_base_radix		},
	{ CALC_GET_MEMORY_INFO,		calc_get_memory_info	},
	{ CALC_MEMORY_FUNCTION,		calc_memory_function	},
	{ CALC_TERMINATE,			calc_terminate			},
	
	{ -1, NULL }
};


/***************************************************************************
function: CALC_FUNCS (argData) - top level function to be used as a ELF macro
*/
elfData
CALC_FUNCS (argData)
elfData argData;
{
	int ix;
	int func_code;
	elfData (*calc_funcP)();
	calc_state *cs;
	int size;
	char *binP;
	elfData state = NULL, status = NULL;
	elfData retData;
	char *mem_handle;
	
	cs = NULL;
	retData = NULL;
	calc_funcP = NULL;

	if ((size = AxBinaryFromArray(argData, 0, &binP)) > 0) {
		memcpy(&mem_handle, binP, size);
		cs = (calc_state *)mem_handle;
	}

	func_code = AxIntFromArray(argData, 1);

	if (calc_func_table[func_code].calc_func_code == func_code)
		calc_funcP = calc_func_table[func_code].calc_funcP;
	else {
		for (ix = 0; calc_func_table[ix].calc_func_code != -1; ix++) {
			if (calc_func_table[ix].calc_func_code == func_code)
				calc_funcP = calc_func_table[ix].calc_funcP;
		}
	}

	if (calc_funcP) {
		status = (*calc_funcP)(&cs, argData);
		mem_handle = (char *)cs;
		if (mem_handle)
			state = AxMakeBinaryData(sizeof(char *), &mem_handle);
		retData = AxMakeArray(2);
		retData = AxAddDataToArray(retData, 0, state);		/* current calc state */
		retData = AxAddDataToArray(retData, 1, status);		/* current func return value */
	}

	return(retData);
}


/***************************************************************************
function: matherr - math error handler
*/
int
matherr (exc)
struct exception *exc;
{
	switch (exc->type) {
	case DOMAIN:
		sprintf(error_msg, "Domain exception (%s)", exc->name);
		break;
	case SING:
		sprintf(error_msg, "Argument singularity (%s)", exc->name);
		break;
	case OVERFLOW:
		sprintf(error_msg, "Overflow range exception (%s)", exc->name);
		break;
	case UNDERFLOW:
		sprintf(error_msg, "Underflow range exception (%s)", exc->name);
		break;
	}
	return(0);
}
